# -*- Python Version: 2.7 -*-
# -*- coding: utf-8 -*-

"""Calculations for Phius Residential Electrical Energy Consumption."""


def cooktop(_num_occupants, _energy_demand):
    # type: (float, float) -> float
    """Return the Phius Cooktop annual energy consumption [kWh] for a single dwelling.

    Assuming a number of meals as per Phius Guidebook V3.02, pg 73 footnote #31
    """
    ANNUAL_MEALS_PER_OCCUPANT = 500
    num_meals = _num_occupants * ANNUAL_MEALS_PER_OCCUPANT
    return _energy_demand * num_meals


def misc_electrical(_num_bedrooms, _floor_area_ft2, _num_dwellings=1):
    # type: (float, float, int) -> float
    """Return Phius Misc. Electrical Loads (MEL) annual energy consumption [kWh] for a single dwelling.

    ### Resnet 2014
    - https://codes.iccsafe.org/content/RESNET3012014P1/4-home-energy-rating-calculation-procedures-
    - Section 4.2.2.5(1): Energy Rating Reference Home
    - kWh = 413 + 0.91 * CFA + 69 * Nbr

    ### Phius Certification Guidebook v24.1.1 | Appendix N | N-7
    - https://www.phius.org/phius-certification-guidebook
    - "The basic protocol for lighting and miscellaneous electric loads is that they are calculated at
    80% of RESNET (2013) levels for the 'Rated Home'."
    - kWh = ((413 * Ndw) + (69 * Nbr) + 0.91 * CFA) * 0.8
    """
    DWELLING_TV_KWH_YR = 413
    BEDROOM_TV_KWH_YR = 69
    MELS_KWH_YR_FT2 = 0.91
    PHIUS_RESNET_FRACTION = 0.8

    a = DWELLING_TV_KWH_YR * _num_dwellings
    b = BEDROOM_TV_KWH_YR * _num_bedrooms
    c = MELS_KWH_YR_FT2 * _floor_area_ft2

    return (a + b + c) * PHIUS_RESNET_FRACTION


def lighting_interior(_floor_area_ft2, _frac_high_efficiency, _num_dwellings=1):
    # type: (float, float, int) -> float
    """Return the Phius Interior Lighting annual energy consumption [kWh] for a single dwelling.

    ### Resnet 2014
    - https://codes.iccsafe.org/content/RESNET3012014P1/4-home-energy-rating-calculation-procedures-
    - Section 4.2.2.5.2.2: Interior Lighting
    - kWh/yr = 0.8 * [(4 - 3 * q_FFIL) / 3.7] * (455 + 0.8 * CFA) + 0.2 * (455 + 0.8 * CFA)

    ### Phius Certification Guidebook v24.1.1 | Appendix N | N-7
    - https://www.phius.org/phius-certification-guidebook
    - "The basic protocol for lighting and miscellaneous electric loads is that they are calculated at
    80% of RESNET (2013) levels for the 'Rated Home'. ... The RESNET lighting formulas have been expressed more
    compactly here but are algebraically equivalent to the published versions."
    - kWh/yr = (0.2 + 0.8 * (4 - 3 * q_FFIL) / 3.7) * ((455 * Ndw) + 0.8 * iCFA) * 0.8
    """

    INT_LIGHTING_W_PER_DWELLING = 455
    INT_LIGHTING_W_FT2 = 0.8
    PHIUS_RESNET_FRACTION = 0.8

    a = 0.2 + 0.8 * (4 - 3 * _frac_high_efficiency) / 3.7
    b = (INT_LIGHTING_W_PER_DWELLING * _num_dwellings) + (INT_LIGHTING_W_FT2 * _floor_area_ft2)

    return a * b * PHIUS_RESNET_FRACTION


def lighting_exterior(_floor_area_ft2, _frac_high_efficiency, _num_dwellings=1):
    # type: (float, float, int) -> float
    """Return the Phius Exterior Lighting annual energy consumption [kWh] for a single dwelling.

    ### Resnet 2014
    - https://codes.iccsafe.org/content/RESNET3012014P1/4-home-energy-rating-calculation-procedures-
    - Section 4.2.2.5.2.3: Exterior Lighting
    - kWh = (100+0.05*FCA)*(1-FF_El)+0.25*(100+0.05*CFA)*FF_EL

    ### Phius Certification Guidebook v24.1.1 | Appendix N | N-7
    - https://www.phius.org/phius-certification-guidebook
    - "The basic protocol for lighting and miscellaneous electric loads is that they are calculated at
    80% of RESNET (2013) levels for the 'Rated Home'. ... The RESNET lighting formulas have been expressed more
    compactly here but are algebraically equivalent to the published versions."
    - kWh/yr = (1 - 0.75 * q_FFIL) * ((100 * Ndw) + (0.05 * iCFA)) * 0.8
    """

    EXT_LIGHTING_KWH_YR_PER_DWELLING = 100
    EXT_LIGHTING_KWH_YR_FT2 = 0.05
    PHIUS_RESNET_FRACTION = 0.8

    a = EXT_LIGHTING_KWH_YR_PER_DWELLING * _num_dwellings
    b = EXT_LIGHTING_KWH_YR_FT2 * _floor_area_ft2
    c = 1 - 0.75 * _frac_high_efficiency

    return c * (a + b) * PHIUS_RESNET_FRACTION


def lighting_garage(_frac_high_efficiency, _num_dwellings=1):
    # type: (float, int) -> float
    """Return the Phius Garage Lighting annual energy consumption [kWh] for a single dwelling .

    ### Resnet 2014
    - https://codes.iccsafe.org/content/RESNET3012014P1/4-home-energy-rating-calculation-procedures-
    - Section 4.2.2.5.1.3: Garage Lighting
    - kWh = 100/dwelling

    ### Phius Certification Guidebook v24.1.1 | Appendix N | N-7
    - https://www.phius.org/phius-certification-guidebook
    - "The basic protocol for lighting and miscellaneous electric loads is that they are calculated at
    80% of RESNET (2013) levels for the 'Rated Home'. ... The RESNET lighting formulas have been expressed more
    compactly here but are algebraically equivalent to the published versions."
    - kWh/yr = 100 * (1 - 0.75 * FFGL) * 0.8

    ### Phius MultiFamily Calculator (v24.0.2) | Nov. 2024
    - kWh/yr = Ndw * (100 * (1 - FFGL) + 25 * FFGL) * 0.8
    """

    GARAGE_LIGHTING_KWH_YR_PER_DWELLING = 100
    PHIUS_RESNET_FRACTION = 0.8
    a = 1 - _frac_high_efficiency
    b = 25 * _frac_high_efficiency

    return _num_dwellings * (GARAGE_LIGHTING_KWH_YR_PER_DWELLING * a + b) * PHIUS_RESNET_FRACTION



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""
AISI S250-21 | “North American Standard for Thermal Transmittance of Building Envelopes With Cold-Formed Steel Framing, 2021 Edition”

https://www.steel.org/2021/09/aisi-publishes-aisi-s250-21-north-american-standard-for-thermal-transmittance-of-building-envelopes-with-cold-formed-steel-framing-2021-edition/

Thermal transmittance (U-factors) of above-grade wall assemblies constructed with standard Cshape studs and track,
where the framing spacing is 6, 12, 16 or 24 inches on-center, shall be determined in accordance with Equation
B3.1.8-1 using Sections B3.1.1 through B3.1.8 or shall be determined in accordance with Appendix 1. Where a
non-standard designation thickness for the cold-formed steel framing member is applicable, values for the OTZ coefficients
from Table B3.1.11 and thermal conductivity from Table B3.1.3-1 shall be based on linear interpolation for the
intermediate value of designation thickness or shall be based on the next larger standard designation thickness.
Extrapolation is not permitted.

NOTE: AISI does not provide an 'SI' version of the coefficients in Table B3.1.11. therefor
ALL values in this module are in imperial units. This includes R-Values, U-Values, and dimensions.
"""

from honeybee_ph_utils.enumerables import CustomEnum


class StudSpacingInches(CustomEnum):
    allowed = [
        "6",
        "12",
        "16",
        "24",
    ]

    def __init__(self, _value=1):
        # type: (str | int) -> None
        super(StudSpacingInches, self).__init__(_value)


class StudThicknessMil(CustomEnum):
    allowed = [
        "33",
        "43",
        "54",
        "68",
    ]

    def __init__(self, _value=1):
        # type: (str | int) -> None
        super(StudThicknessMil, self).__init__(_value)


# [AISI S250-21w/S1-22] Table B3.1.1-1 | OTZ Coefficients
OTZ_COEFFICIENTS = {
    StudSpacingInches("6"): {
        StudThicknessMil("33"): [1.8583, 0.07478, 0.1488, -0.001859, -0.005103, 0.002013],
        StudThicknessMil("43"): [1.9826, 0.0736, 0.1501, -0.001816, -0.005314, 0.002149],
        StudThicknessMil("54"): [2.0814, 0.07131, 0.1522, -0.001713, -0.005295, 0.00205],
        StudThicknessMil("68"): [2.211, 0.06816, 0.1508, -0.001652, -0.005576, 0.0023],
    },
    StudSpacingInches("12"): {
        StudThicknessMil("33"): [2.1584, 0.05118, 0.2079, -0.001384, -0.005367, 0.002253],
        StudThicknessMil("43"): [2.2077, 0.06381, 0.1992, -0.001713, -0.006235, 0.003499],
        StudThicknessMil("54"): [2.2974, 0.06439, 0.2043, -0.001686, -0.006908, 0.003943],
        StudThicknessMil("68"): [2.4136, 0.05185, 0.2166, -0.001216, -0.00684, 0.003748],
    },
    StudSpacingInches("16"): {
        StudThicknessMil("33"): [2.2771, 0.03843, 0.1964, -0.001141, -0.005237, 0.003197],
        StudThicknessMil("43"): [2.3769, 0.04037, 0.2011, -0.001195, -0.005677, 0.003714],
        StudThicknessMil("54"): [2.4945, 0.04089, 0.1996, -0.001161, -0.005719, 0.003927],
        StudThicknessMil("68"): [2.5917, 0.04614, 0.1922, -0.001391, -0.005884, 0.004606],
    },
    StudSpacingInches("24"): {
        StudThicknessMil("33"): [3.182, -0.02946, 0.2432, 0, -0.00752, 0.003572],
        StudThicknessMil("43"): [2.751, 0.0128, 0.1965, -0.00074, -0.006709, 0.005169],
        StudThicknessMil("54"): [2.572, 0.00426, 0.2285, 0, -0.0061, 0.003509],
        StudThicknessMil("68"): [2.936, -0.00324, 0.2256, 0, -0.00643, 0.00419],
    },
}  # type: dict[StudSpacingInches, dict[StudThicknessMil, list[float]]]


# [AISI S250-21w/S1-22] Table B3.1.3-1 | Cold-Formed Steel Thermal Conductivity (Btu/hr-ft2-F)
STEEL_CONDUCTIVITY = {
    StudThicknessMil("33"): 381.0,
    StudThicknessMil("43"): 495.0,
    StudThicknessMil("54"): 622.0,
    StudThicknessMil("68"): 783.0,
}


def overall_thermal_zone(
    c_0,
    c_1,
    c_2,
    c_3,
    c_4,
    c_5,
    r_cavity,
    r_ext_insulation,
):
    # type: (float, float, float, float, float, float, float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.1 Overall Thermal Zone.

    The overall thermal zone (OTZ) shall be determined in accordance with Equation B3.1.1-1.
    The appropriate coefficients for use in Equation B3.1.1-1 shall be obtained from Table B3.1.11
    with the given on-center spacing of framing members and C-shape designation thickness.

        Eq. B3.1.1-1: OTZ = C_0 + C_1 * R_cav + C_2 * R_she + C_3 * R_cav^2 + C_4 * R_she^2 + C_5 * R_cav * R_she

    r_cavity (r_cav) = R_air + R_ins
        * R_ins: Thermal resistance (R-value) of cavity insulation (h-ft2-F/Btu)
            * = 0 where wall cavity contains no insulation
        * R_air: Thermal resistance (R-value) of the cavity air space (h-ft2-F/Btu)
            * = 0.91 where wall cavity contains an air space
            * = 0 where wall cavity contains no air space

    Args:
        C_0: Coefficient from Table B3.1.1-1
        C_1: Coefficient from Table B3.1.1-1
        C_2: Coefficient from Table B3.1.1-1
        C_3: Coefficient from Table B3.1.1-1
        C_4: Coefficient from Table B3.1.1-1
        C-5: Coefficient from Table B3.1.1-1
        r_cavity (R_cav): Total thermal resistance (R-value) of the wall cavity (h-ft2-F/Btu)
        r_ext_insulation (R_she): Thermal resistance (R-value) of exterior continuous insulation (hr-ft2-F/Btu). R_she does
            not include wood or gypsum panels.
    Returns:
        OTZ: Overall thermal zone (inch)
    """

    return (
        c_0
        + c_1 * r_cavity
        + c_2 * r_ext_insulation
        + c_3 * r_cavity**2
        + c_4 * r_ext_insulation**2
        + c_5 * r_cavity * r_ext_insulation
    )


def framing_factor(stud_flange_width_inch, stud_thickness_mils):
    # type: (float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.2 C-Shape Framing Factor.

    The C-shape framing factor (FFcs) for the cold-formed steel framing shall be determined in
    accordance with Equation B3.1.2-1:

        Eq. B3.1.2-1: FF_cs = (CFS_t / 1000) / CFS_flange

    Args:
        stud_thickness_mils (CFS_t): Designation thickness of cold-formed steel framing member (mils)
        stud_flange_width_inch (CFS_flange): Flange width of cold-formed steel framing member (inches)
    Returns:
        float: (FF_cs) C-shape framing factor (unitless)
    """

    return ((stud_thickness_mils) / 1000) / stud_flange_width_inch


def stud_web_r_value(stud_depth_inch, conductivity):
    # type: (float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.3.1 Thermal Resistance of C-Shape Web.

    The thermal resistance (R-value) of the web of the cold-formed steel framing member (Rs-  wall )
    shall be determined in accordance with Equation B3.1.3-1:

        * R_s-wall = CFS_depth / k

    Args:
        stud_depth (CFS_depth): Web depth of cold-formed steel framing member that is in contact with cavity insulation (inches)
            as illustrated by Figure B3.1.3-1 without cavity air space and Figure B3.1.3-2 with cavity air space
        conductivity (k): Thermal conductivity of cold-formed steel as selected from Table B3.1.3-1 (Btu/h-ft2-F)
    Returns:
        float: (R_s-wall) Thermal resistance (R-value) of the web of the cold-formed steel framing member (h-ft2-F/Btu)
    """

    return stud_depth_inch / conductivity


def u_value_at_stud(framing_factor, r_value_cavity_insulation, r_value_stud_web):
    # type: (float, float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.3.2 Thermal Transmittance of Cavity with Insulation.

    The parallel path thermal transmittance (U-factor) of the assembly where insulation(s) are in contact with
    cold-formed steel framing member (U1 ) shall be determined in accordance with Equation B3.1.3-2. Where there is
    no insulation in the wall cavity, Equation B3.1.3-2 shall not be used and the resultant value of R 1 in
    Equation B3.1.3-3 shall be “zero” (0).

        * U_1 = (1 - FFcs) / R_ins + FFcs / Rs-wall

    Args:
        framing_factor (FF_cs): C-shape framing factor determined from Equation B3.1.2-1
        r_value_cavity_insulation (R_ins): Thermal resistance (R-value) of cavity insulation (h-ft2-F/Btu)
        r_value_stud_web (R_s-wall): Thermal resistance (R-value) of the wall assembly at the web of the cold-formed steel framing
            member determined from Equation B3.1.3-1 (h-ft2-F/Btu)
    Returns:
        float: (U_1) Parallel path thermal transmittance (U-factor) of the assembly where insulation(s) are
            in contact with cold-formed steel framing member (Btu/hr-ft2-F)
    """

    return (1 - framing_factor) / r_value_cavity_insulation + framing_factor / r_value_stud_web


def r_value_series_stud_path(
    r_ext_cladding,
    r_ext_insulation,
    r_ext_sheathing,
    r_3,
    r_int_sheathing,
    r_se=0.17,
    r_si=0.68,
):
    # type: (float, float, float, float, float, float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.6.1 Thermal Resistance (R-value) of Series Paths | C-Shape Path.

    The effective thermal resistance (R-value) of all materials in the design assembly in series for the
    path containing cold-formed steel (R_sps), illustrated as the C-shape path in Figure B3.1.3-1 or
    Figure B3.1.3-2, shall be determined using Equation B3.1.6-1:

        * R_sps = R_o + Side_ext + R_she + Sheath_ext + R_3 + Sheath_int + R_i

    Args:
        r_ext_cladding (Side_ext): Thermal resistance (R-value) of exterior siding (h-ft2-F/Btu)
        r_ext_insulation (R_she): Thermal resistance (R-value) of exterior continuous insulation (h-ft2-F/Btu)
        r_ext_sheathing (Sheath_ext): Thermal resistance (R-value) of all exterior sheathing(s) (h-ft2-F/Btu)
        R_3: Thermal resistance (R-value) determined from Equation B3.1.5-1 (h-ft2-F/Btu)
        r_int_sheathing (Sheath_int): Thermal resistance (R-value) of all interior sheathing(s) (h-ft2-F/Btu)
        r_se (R_o): [Default=0.17] Thermal resistance (R-value) of outside air film (h-ft2-F/Btu)
        r_si (R_i): [Default=0.68] Thermal resistance (R-value) of inside air film (h-ft2-F/Btu)
    Returns:
        float: (R_sps) Effective thermal resistance (R-value) of all materials in the design assembly
            in series for the path containing cold-formed steel (h-ft2-F/Btu)
    """
    return r_se + r_ext_cladding + r_ext_insulation + r_ext_sheathing + r_3 + r_int_sheathing + r_si


def r_value_series_cavity_path(
    r_ext_cladding,
    r_ext_insulation,
    r_ext_sheathing,
    r_cavity_insulation,
    r_int_sheathing,
    r_se=0.17,
    r_si=0.68,
):
    # type: (float, float, float, float, float, float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.6.1 Thermal Resistance (R-value) of Series Paths | Cavity Path.

    B3.1.6.2 Cavity Path The thermal resistance (R-value) of all materials in the design assembly in series for the path
    containing only cavity insulation (R_spc), illustrated as the cavity path in Figure B3.1.31 or Figure B3.1.3-2,
    shall be determined using Equation B3.1.6-2:

        * R_spc = R_o + Side_ext + R_she + Sheath_ext + R_ins + R_air + Sheath_int + R_i

    Args:
        R_o: Thermal resistance (R-value) of outside air film (h-ft2F/Btu)
        r_ext_cladding (Side_ext): Thermal resistance (R-value) of exterior siding (h-ft2F/Btu)
        r_ext_insulation (R_she): Thermal resistance (R-value) of exterior continuous insulation (h-ft2F/Btu)
        r_ext_sheathing (Sheath_ext): Thermal resistance (R-value) of all exterior sheathing(s) (h-ft2F/Btu)
        r_cavity_insulation (R_ins): Thermal resistance (R-value) of cavity insulation (h-ft2F/Btu)
        R_air: Thermal resistance (R-value) of the cavity air space (h-ft2F/Btu)
           : 0.91 where wall cavity contains an air space
           : 0.00 where wall cavity contains no air space
        r_int_sheathing (Sheath_int): Thermal resistance (R-value) of all interior sheathing(s) (h-ft2F/Btu)
        R_i: Thermal resistance (R-value) of inside air film (h-ft2F/Btu)
    Returns:
        float: (R_spc) Effective thermal resistance (R-value) of all materials in the design assembly
            in series for the path containing only cavity insulation (h-ft2-F/Btu)
    """

    return r_se + r_ext_cladding + r_ext_insulation + r_ext_sheathing + r_cavity_insulation + r_int_sheathing + r_si


def overall_thermal_zone_framing_factor(_overall_thermal_zone, _stud_spacing_inch):
    # type: (float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.7 OTZ Framing Factor.

    The OTZ Framing Factor (FF_otz) shall be determined using Equation B3.1.7-1:
        * FF_otz = OTZ / FS

    Args:
        _overall_thermal_zone (OTZ): Overall thermal zone (inch)
        _stud_spacing_inch (fs): Spacing of cold-formed steel framing members (inches)
    Returns:
        float: (FF_otz) OTZ Framing Factor (unitless)
    """

    return _overall_thermal_zone / _stud_spacing_inch


def u_value_total(_overall_thermal_zone_framing_factor, _r_value_series_cavity_path, _r_value_series_stud_path):
    # type: (float, float, float) -> float
    """
    [AISI S250-21w/S1-22] B3.1.8 Overall Thermal Transmittance (U-Factor).

    The overall thermal transmittance (U-factor) of the above-grade wall assembly (U_o)
    shall be determined using Equation B3.1.8-1:

        * U_o = (1 - FF_otz) / R_spc + FF_otz / R_sps

    Args:
        _overall_thermal_zone_framing_factor (FF_otz) : OTZ Framing Factor determined from Equation B3.1.7-1 (dimensionless)
        _r_value_series_cavity_path (R_spc) : Thermal resistance (R-value) of all materials in the design assembly
            in series for the cavity path determined from Equation B3.1.6-2 (h-ft2-F/Btu)
        _r_value_series_stud_path (R_sps) : Thermal resistance (R-value) of all materials in the design assembly
            in series for the C-shape path determined from Equation B3.1.6-1 (h-ft2-F/Btu)
    Returns:
        float: (U_o) Overall thermal transmittance (U-factor) of the above-grade wall assembly (Btu/hr-ft2-F)
    """
    return (
        1 - _overall_thermal_zone_framing_factor
    ) / _r_value_series_cavity_path + _overall_thermal_zone_framing_factor / _r_value_series_stud_path


def calculate_stud_cavity_effective_u_value(
    _r_ext_cladding,
    _r_ext_insulation,
    _r_ext_sheathing,
    _r_cavity_insulation,
    _stud_spacing_inch,
    _stud_thickness_mil,
    _stud_flange_width_inch,
    _stud_depth_inch,
    _steel_conductivity,
    _r_int_sheathing,
    _r_se=0.17,
    _r_si=0.68,
):
    # type: (float, float, float, float, StudSpacingInches, StudThicknessMil, float, float, float, float, float, float) -> float
    try:
        C0, C1, C2, C3, C4, C5 = OTZ_COEFFICIENTS[_stud_spacing_inch][_stud_thickness_mil]
    except KeyError as e:
        msg = "The given stud spacing: '{}' and stud-thickness '{}' is not supported. ".format(
            _stud_spacing_inch.value, _stud_thickness_mil.value
        )
        msg += "Please use one of the following combinations:\n"
        msg += "\tStud Spacing: {}\n".format(", ".join(StudSpacingInches.allowed))
        msg += "\tStud Thickness: {}\n".format(", ".join(StudThicknessMil.allowed))
        msg += "Error: {}".format(e)
        raise KeyError(msg)

    otz = overall_thermal_zone(C0, C1, C2, C3, C4, C5, _r_cavity_insulation, _r_ext_insulation)
    ff = framing_factor(_stud_flange_width_inch, float(_stud_thickness_mil.value))
    web_r_value = stud_web_r_value(_stud_depth_inch, _steel_conductivity)
    stud_u_value = u_value_at_stud(ff, _r_cavity_insulation, web_r_value)

    r_1 = 1 / stud_u_value
    r_2 = 0  #  <----- TODO: Calculate this properly calc U_2
    r_3 = r_1 + r_2

    r_stud_path = r_value_series_stud_path(
        _r_ext_cladding, _r_ext_insulation, _r_ext_sheathing, r_3, _r_int_sheathing, _r_se, _r_si
    )
    r_cavity_path = r_value_series_cavity_path(
        _r_ext_cladding, _r_ext_insulation, _r_ext_sheathing, _r_cavity_insulation, _r_int_sheathing, _r_se, _r_si
    )
    otz_ff = overall_thermal_zone_framing_factor(otz, float(_stud_spacing_inch.value))

    u_overall = u_value_total(otz_ff, r_cavity_path, r_stud_path)
    r_overall = 1 / u_overall

    # Get the stud cavity U-Value by itself
    r_stud_cavity = (
        r_overall - _r_ext_cladding - _r_ext_insulation - _r_ext_sheathing - _r_int_sheathing - _r_se - _r_si
    )
    u_stud_cavity = 1 / r_stud_cavity
    return u_stud_cavity



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Passive-House Style Monthly Climate Data"""

from copy import copy

try:
    from itertools import izip as zip  # type: ignore
except ImportError:
    pass  # Python3

try:
    from typing import Any, Collection, Dict, List, Optional, Union
except ImportError:
    pass  # IronPython 2.7

from honeybee_ph import _base


class Climate_MonthlyValueSet(_base._Base):
    """A set of 12 monthly values (temp, radiation, etc)."""

    january = 0.0
    february = 0.0
    march = 0.0
    april = 0.0
    may = 0.0
    june = 0.0
    july = 0.0
    august = 0.0
    september = 0.0
    october = 0.0
    november = 0.0
    december = 0.0
    months = [
        "january",
        "february",
        "march",
        "april",
        "may",
        "june",
        "july",
        "august",
        "september",
        "october",
        "november",
        "december",
    ]

    def __init__(self, _values=[0.0] * 12):
        # type: (Collection[float]) -> None
        super(Climate_MonthlyValueSet, self).__init__()
        self.values = _values

    @property
    def values(self):
        # type: () -> List[float]
        return [getattr(self, month) for month in self.months]

    @values.setter
    def values(self, _in):
        # type: (Collection[float]) -> None
        if (_in is None) or (len(_in) != 12):
            msg = "Error: Monthly data should be a collection of 12 numeric items.\n" "Got a {} of length: {}?".format(
                type(_in), len(_in)
            )
            raise Exception(msg)

        for val, month_name in zip(_in, self.months):
            setattr(self, month_name, val)

    def __copy__(self):
        # type: () -> Climate_MonthlyValueSet
        obj = Climate_MonthlyValueSet(copy(self.values))
        obj.set_base_attrs_from_source(self)
        for month in self.months:
            setattr(obj, month, getattr(self, month))

        return obj

    def duplicate(self):
        # type: () -> Climate_MonthlyValueSet
        return self.__copy__()

    def to_dict(self):
        # type: () -> Dict[str, float]
        d = {}

        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        for month in self.months:
            d[month] = getattr(self, month)

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, float]) -> Climate_MonthlyValueSet
        obj = cls()
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.user_data = _input_dict.get("user_data", {})

        for month in cls.months:
            setattr(obj, month, _input_dict.get(month))

        return obj


class Climate_MonthlyTempCollection(_base._Base):
    """Collection class to organize monthly temperature values"""

    def __init__(
        self,
        _air=Climate_MonthlyValueSet(),
        _dewpoint=Climate_MonthlyValueSet(),
        _sky=Climate_MonthlyValueSet(),
        _ground=Climate_MonthlyValueSet(),
    ):
        # type: (Climate_MonthlyValueSet, Climate_MonthlyValueSet, Climate_MonthlyValueSet, Climate_MonthlyValueSet) -> None
        super(Climate_MonthlyTempCollection, self).__init__()
        self.air_temps = _air
        self.dewpoints = _dewpoint
        self.sky_temps = _sky
        self.ground_temps = _ground

    def __copy__(self):
        # type: () -> Climate_MonthlyTempCollection
        obj = Climate_MonthlyTempCollection(
            self.air_temps.duplicate(),
            self.dewpoints.duplicate(),
            self.sky_temps.duplicate(),
            self.ground_temps.duplicate(),
        )
        obj.set_base_attrs_from_source(self)

        return obj

    def duplicate(self):
        # type: () -> Climate_MonthlyTempCollection
        return self.__copy__()

    def to_dict(self):
        # type: () -> Dict[str, Dict[str, float]]
        d = {}

        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        d["air_temps"] = self.air_temps.to_dict()
        d["dewpoints"] = self.dewpoints.to_dict()
        d["sky_temps"] = self.sky_temps.to_dict()
        d["ground_temps"] = self.ground_temps.to_dict()

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Dict[str, float]]) -> Climate_MonthlyTempCollection
        obj = cls(
            _air=Climate_MonthlyValueSet.from_dict(_input_dict["air_temps"]),
            _dewpoint=Climate_MonthlyValueSet.from_dict(_input_dict["dewpoints"]),
            _sky=Climate_MonthlyValueSet.from_dict(_input_dict["sky_temps"]),
            _ground=Climate_MonthlyValueSet.from_dict(_input_dict["ground_temps"]),
        )
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.user_data = _input_dict.get("user_data", {})

        return obj


class Climate_MonthlyRadiationCollection(_base._Base):

    def __init__(
        self,
        _north=Climate_MonthlyValueSet(),
        _east=Climate_MonthlyValueSet(),
        _south=Climate_MonthlyValueSet(),
        _west=Climate_MonthlyValueSet(),
        _glob=Climate_MonthlyValueSet(),
    ):
        super(Climate_MonthlyRadiationCollection, self).__init__()
        # type: (Climate_MonthlyValueSet, Climate_MonthlyValueSet, Climate_MonthlyValueSet, Climate_MonthlyValueSet) -> None
        self.north = _north
        self.east = _east
        self.south = _south
        self.west = _west
        self.glob = _glob

    def __copy__(self):
        # type: () -> Climate_MonthlyRadiationCollection
        obj = Climate_MonthlyRadiationCollection(
            self.north.duplicate(),
            self.east.duplicate(),
            self.south.duplicate(),
            self.west.duplicate(),
            self.glob.duplicate(),
        )
        obj.set_base_attrs_from_source(self)

        return obj

    def duplicate(self):
        # type: () -> Climate_MonthlyRadiationCollection
        return self.__copy__()

    def to_dict(self):
        # type: () -> Dict[str, Dict[str, float]]
        d = {}
        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        d["north"] = self.north.to_dict()
        d["east"] = self.east.to_dict()
        d["south"] = self.south.to_dict()
        d["west"] = self.west.to_dict()
        d["glob"] = self.glob.to_dict()

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Dict[str, float]]) -> Climate_MonthlyRadiationCollection
        obj = cls(
            _north=Climate_MonthlyValueSet.from_dict(_input_dict["north"]),
            _east=Climate_MonthlyValueSet.from_dict(_input_dict["east"]),
            _south=Climate_MonthlyValueSet.from_dict(_input_dict["south"]),
            _west=Climate_MonthlyValueSet.from_dict(_input_dict["west"]),
            _glob=Climate_MonthlyValueSet.from_dict(_input_dict["glob"]),
        )
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.user_data = _input_dict.get("user_data", {})

        return obj


class Climate_PeakLoadValueSet(_base._Base):
    """A set of Peak Load data."""

    def __init__(
        self,
        _temp=0.0,
        _rad_north=0.0,
        _rad_east=0.0,
        _rad_south=0.0,
        _rad_west=0.0,
        _rad_global=0.0,
        _dewpoint_temp=None,
        _sky_temp=None,
        _ground_temp=None,
    ):
        # type: (float, float, float, float, float, float, Optional[float], Optional[float], Optional[float]) -> None
        super(Climate_PeakLoadValueSet, self).__init__()
        self.temp = _temp
        self.rad_north = _rad_north
        self.rad_east = _rad_east
        self.rad_south = _rad_south
        self.rad_west = _rad_west
        self.rad_global = _rad_global
        self.dewpoint = _dewpoint_temp
        self.sky_temp = _sky_temp
        self.ground_temp = _ground_temp

    def __copy__(self):
        # type: () -> Climate_PeakLoadValueSet
        obj = Climate_PeakLoadValueSet(
            self.temp,
            self.rad_north,
            self.rad_east,
            self.rad_south,
            self.rad_west,
            self.rad_global,
            self.dewpoint,
            self.sky_temp,
            self.ground_temp,
        )
        obj.set_base_attrs_from_source(self)

        return obj

    def duplicate(self):
        # type: () -> Climate_PeakLoadValueSet
        return self.__copy__()

    def to_dict(self):
        # type: () -> Dict[str, float]
        d = {}

        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        d["temp"] = self.temp
        d["rad_north"] = self.rad_north
        d["rad_east"] = self.rad_east
        d["rad_south"] = self.rad_south
        d["rad_west"] = self.rad_west
        d["rad_global"] = self.rad_global
        d["dewpoint"] = self.dewpoint
        d["sky_temp"] = self.sky_temp
        d["ground_temp"] = self.ground_temp

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, float]) -> Climate_PeakLoadValueSet
        obj = cls()

        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.user_data = _input_dict.get("user_data", {})

        obj.temp = _input_dict["temp"]
        obj.rad_north = _input_dict["rad_north"]
        obj.rad_east = _input_dict["rad_east"]
        obj.rad_south = _input_dict["rad_south"]
        obj.rad_west = _input_dict["rad_west"]
        obj.rad_global = _input_dict["rad_global"]
        obj.dewpoint = _input_dict["dewpoint"]
        obj.sky_temp = _input_dict["sky_temp"]
        obj.ground_temp = _input_dict["ground_temp"]

        return obj


class Climate_PeakLoadCollection(_base._Base):
    """A Collection of Peak Loads (Heating and Cooling)."""

    def __init__(
        self,
        _heat_load_1=Climate_PeakLoadValueSet(),
        _heat_load_2=Climate_PeakLoadValueSet(),
        _cooling_load_1=Climate_PeakLoadValueSet(),
        _cooling_load_2=Climate_PeakLoadValueSet(),
    ):
        # type: (Climate_PeakLoadValueSet, Climate_PeakLoadValueSet, Climate_PeakLoadValueSet, Climate_PeakLoadValueSet) -> None
        super(Climate_PeakLoadCollection, self).__init__()
        self.heat_load_1 = _heat_load_1
        self.heat_load_2 = _heat_load_2
        self.cooling_load_1 = _cooling_load_1
        self.cooling_load_2 = _cooling_load_2

    def __copy__(self):
        # type: () -> Climate_PeakLoadCollection
        obj = Climate_PeakLoadCollection(
            self.heat_load_1.duplicate(),
            self.heat_load_2.duplicate(),
            self.cooling_load_1.duplicate(),
            self.cooling_load_2.duplicate(),
        )
        obj.set_base_attrs_from_source(self)

        return obj

    def duplicate(self):
        # type: () -> Climate_PeakLoadCollection
        return self.__copy__()

    def to_dict(self):
        # type: () -> Dict[str, Union[Dict[str, float], str]]
        d = {}

        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        d["heat_load_1"] = self.heat_load_1.to_dict()
        d["heat_load_2"] = self.heat_load_2.to_dict()
        d["cooling_load_1"] = self.cooling_load_1.to_dict()
        d["cooling_load_2"] = self.cooling_load_2.to_dict()
        d["display_name"] = self.display_name

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> Climate_PeakLoadCollection
        obj = cls(
            Climate_PeakLoadValueSet.from_dict(_input_dict["heat_load_1"]),
            Climate_PeakLoadValueSet.from_dict(_input_dict["heat_load_2"]),
            Climate_PeakLoadValueSet.from_dict(_input_dict["cooling_load_1"]),
            Climate_PeakLoadValueSet.from_dict(_input_dict["cooling_load_2"]),
        )
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.user_data = _input_dict.get("user_data", {})

        return obj


class Climate_Ground(_base._Base):
    def __init__(self):
        # type: () -> None
        super(Climate_Ground, self).__init__()
        self.ground_thermal_conductivity = 2
        self.ground_heat_capacity = 1000
        self.ground_density = 2000
        self.depth_groundwater = 3
        self.flow_rate_groundwater = 0.05

    def __copy__(self):
        # type: () -> Climate_Ground
        obj = Climate_Ground()
        obj.set_base_attrs_from_source(self)
        for attr_nm, attr_val in vars(self).items():
            setattr(obj, attr_nm, attr_val)

        return obj

    def duplicate(self):
        # type: () -> Climate_Ground
        return self.__copy__()

    def to_dict(self):
        # type: () -> Dict
        d = {}

        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        d["ground_thermal_conductivity"] = self.ground_thermal_conductivity
        d["ground_heat_capacity"] = self.ground_heat_capacity
        d["ground_density"] = self.ground_density
        d["depth_groundwater"] = self.depth_groundwater
        d["flow_rate_groundwater"] = self.flow_rate_groundwater

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> Climate_Ground
        obj = cls()
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.user_data = _input_dict.get("user_data", {})

        obj.ground_thermal_conductivity = _input_dict.get(
            "ground_thermal_conductivity", obj.ground_thermal_conductivity
        )
        obj.ground_heat_capacity = _input_dict.get("ground_heat_capacity", obj.ground_heat_capacity)
        obj.ground_density = _input_dict.get("ground_density", obj.ground_density)
        obj.depth_groundwater = _input_dict.get("depth_groundwater", obj.depth_groundwater)
        obj.flow_rate_groundwater = _input_dict.get("flow_rate_groundwater", obj.flow_rate_groundwater)

        return obj


class Climate(_base._Base):
    def __init__(
        self,
        _display_name="New York",
        _station_elevation=0.0,
        _daily_temp_swing=8.0,
        _average_wind_speed=4.0,
        _monthly_temps=Climate_MonthlyTempCollection(),
        _monthly_radiation=Climate_MonthlyRadiationCollection(),
        _peak_loads=Climate_PeakLoadCollection(),
    ):
        # type: (str, float, float, float, Climate_MonthlyTempCollection, Climate_MonthlyRadiationCollection, Climate_PeakLoadCollection) -> None
        super(Climate, self).__init__()
        self.display_name = _display_name
        self.station_elevation = _station_elevation  # m
        self.summer_daily_temperature_swing = _daily_temp_swing  # Deg-K
        self.average_wind_speed = _average_wind_speed  # m/s

        self.ground = Climate_Ground()
        self.monthly_temps = _monthly_temps
        self.monthly_radiation = _monthly_radiation
        self.peak_loads = _peak_loads

    def to_dict(self):
        # type: () -> Dict
        d = {}

        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        d["station_elevation"] = self.station_elevation
        d["summer_daily_temperature_swing"] = self.summer_daily_temperature_swing
        d["average_wind_speed"] = self.average_wind_speed
        d["ground"] = self.ground.to_dict()
        d["monthly_temps"] = self.monthly_temps.to_dict()
        d["monthly_radiation"] = self.monthly_radiation.to_dict()
        d["peak_loads"] = self.peak_loads.to_dict()

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> Climate
        obj = cls()

        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.user_data = _input_dict.get("user_data", {})

        obj.station_elevation = _input_dict["station_elevation"]
        obj.summer_daily_temperature_swing = _input_dict["summer_daily_temperature_swing"]
        obj.average_wind_speed = _input_dict["average_wind_speed"]

        obj.ground = Climate_Ground.from_dict(_input_dict.get("ground", {}))
        obj.monthly_temps = Climate_MonthlyTempCollection.from_dict(_input_dict.get("monthly_temps", {}))
        obj.monthly_radiation = Climate_MonthlyRadiationCollection.from_dict(_input_dict.get("monthly_radiation", {}))
        obj.peak_loads = Climate_PeakLoadCollection.from_dict(_input_dict.get("peak_loads", {}))

        return obj

    def __copy__(self):
        # type: () -> Climate
        obj = Climate()
        obj.set_base_attrs_from_source(self)
        for attr_nm, attr_val in vars(self).items():
            setattr(obj, attr_nm, attr_val)

        return obj

    def duplicate(self):
        # type: () -> Climate
        return self.__copy__()


class Location(_base._Base):
    """Geographic Location Information."""

    def __init__(
        self,
        latitude=40.6,
        longitude=-73.8,
        site_elevation=None,
        climate_zone=1,
        hours_from_UTC=-4,
    ):
        # type: (float, float, Optional[float], int, int) -> None
        super(Location, self).__init__()
        self.latitude = latitude
        self.longitude = longitude
        self.site_elevation = site_elevation
        self.climate_zone = climate_zone
        self.hours_from_UTC = hours_from_UTC

    def to_dict(self):
        # type: () -> Dict
        d = {}

        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        d["latitude"] = self.latitude
        d["longitude"] = self.longitude
        d["site_elevation"] = self.site_elevation
        d["climate_zone"] = self.climate_zone
        d["hours_from_UTC"] = self.hours_from_UTC

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> Location
        obj = cls(
            _input_dict["latitude"],
            _input_dict["longitude"],
            _input_dict["site_elevation"],
            _input_dict["climate_zone"],
            _input_dict["hours_from_UTC"],
        )
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.user_data = _input_dict.get("user_data", {})

        return obj

    def __copy__(self):
        # type: () -> Location
        obj = Location(
            self.latitude,
            self.longitude,
            self.site_elevation,
            self.climate_zone,
            self.hours_from_UTC,
        )
        obj.set_base_attrs_from_source(self)

        return obj

    def duplicate(self):
        # type: () -> Location
        return self.__copy__()


class PHPPCodes(_base._Base):
    """Settings / names if using Pre-loaded PHPP Library Data"""

    def __init__(
        self,
        _country_code="US-United States of America",
        _region_code="New York",
        _dataset_name="US0055c-New York",
    ):
        # type: (str, str, str) -> None
        super(PHPPCodes, self).__init__()
        self.country_code = _country_code
        self.region_code = _region_code
        self.dataset_name = _dataset_name

    def to_dict(self):
        # type: () -> Dict
        d = {}

        d["country_code"] = self.country_code
        d["region_code"] = self.region_code
        d["dataset_name"] = self.dataset_name
        d["display_name"] = self.dataset_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> PHPPCodes
        obj = cls(
            _input_dict["country_code"],
            _input_dict["region_code"],
            _input_dict["dataset_name"],
        )
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.user_data = _input_dict.get("user_data", {})

        return obj

    def __copy__(self):
        # type: () -> PHPPCodes
        obj = PHPPCodes(
            self.country_code,
            self.region_code,
            self.dataset_name,
        )
        obj.set_base_attrs_from_source(self)

        return obj

    def duplicate(self):
        # type: () -> PHPPCodes
        return self.__copy__()


class Site(_base._Base):
    """Location and Climate data for the building site."""

    def __init__(
        self,
        _location=Location(),
        _climate=Climate(),
        _phpp_library_codes=PHPPCodes(),
    ):
        # type: (Location, Climate, PHPPCodes) -> None
        super(Site, self).__init__()
        self.location = _location
        self.climate = _climate
        self.phpp_library_codes = _phpp_library_codes

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}

        d["location"] = self.location.to_dict()
        d["climate"] = self.climate.to_dict()
        d["phpp_library_codes"] = self.phpp_library_codes.to_dict()
        d["display_name"] = self.display_name
        d["identifier"] = self.identifier
        d["user_data"] = copy(self.user_data)

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict[str, Any]) -> Site
        obj = cls(
            Location.from_dict(_input_dict["location"]),
            Climate.from_dict(_input_dict["climate"]),
            PHPPCodes.from_dict(_input_dict["phpp_library_codes"]),
        )
        obj.display_name = _input_dict.get("display_name", obj.display_name)
        obj.identifier = _input_dict.get("identifier", obj.identifier)
        obj.user_data = _input_dict.get("user_data", {})

        return obj

    def __copy__(self):
        # type: () -> Site
        obj = Site()

        obj.set_base_attrs_from_source(self)
        obj.location = self.location.duplicate()
        obj.climate = self.climate.duplicate()
        obj.phpp_library_codes = self.phpp_library_codes.duplicate()

        return obj

    def duplicate(self):
        # type: () -> Site
        return self.__copy__()



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Passive House properties for honeybee_energy.construction.opaque.OpaqueConstruction Objects"""

try:
    from typing import Any
except ImportError:
    pass  # Python 2.7


class OpaqueConstructionPhProperties_FromDictError(Exception):
    def __init__(self, _expected_types, _input_type):
        self.msg = 'Error: Expected type of "{}". Got: {}'.format(_expected_types, _input_type)
        super(OpaqueConstructionPhProperties_FromDictError, self).__init__(self.msg)


class OpaqueConstructionPhProperties(object):
    def __init__(self, _host=None):
        # type (Any) -> None
        self._host = _host
        self.id_num = 0

    @property
    def host(self):
        return self._host

    def duplicate(self, new_host=None):
        # type: (Any) -> OpaqueConstructionPhProperties
        return self.__copy__(new_host)

    def __copy__(self, new_host=None):
        # type: (Any) -> OpaqueConstructionPhProperties
        host = new_host or self.host

        new_obj = self.__class__(host)
        new_obj.id_num = self.id_num
        return new_obj

    def to_dict(self, abridged=False):
        # type: (bool) -> dict
        d = {}

        if abridged:
            d["type"] = "OpaqueConstructionPhPropertiesAbridged"
        else:
            d["type"] = "OpaqueConstructionPhProperties"

        d["id_num"] = self.id_num
        return {"ph": d}

    @classmethod
    def from_dict(cls, _input_dict, host):
        # type: (dict, Any) -> OpaqueConstructionPhProperties
        valid_types = (
            "OpaqueConstructionPhProperties",
            "OpaqueConstructionPhPropertiesAbridged",
        )
        if _input_dict["type"] not in valid_types:
            raise OpaqueConstructionPhProperties_FromDictError(valid_types, _input_dict["type"])

        new_obj = cls(host)
        new_obj.id_num = _input_dict["id_num"]
        return new_obj

    def __str__(self):
        return "{}(id_num={!r})".format(self.__class__.__name__, self.id_num)

    def __repr__(self):
        return str(self)

    def ToString(self):
        return str(self)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC: Hot Water Piping."""

from copy import copy
from math import radians

try:
    from typing import Any, Dict, List, Optional, Union
except ImportError:
    pass  # IronPython

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.polyline import LineSegment3D
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry", e)

try:
    from honeybee_phhvac import _base
except ImportError as e:
    raise ImportError("Failed to import honeybee_phhvac", e)

try:
    from honeybee_ph_utils import enumerables
except ImportError as e:
    raise ImportError("Failed to import honeybee_ph_utils", e)


# -- Piping  Enums ------------------------------------------------------------


class PhHvacPipeMaterial(enumerables.CustomEnum):
    allowed = [
        "1-COPPER_M",
        "2-COPPER_L",
        "3-COPPER_K",
        "4-CPVC_CTS_SDR",
        "5-CPVC_SCH_40",
        "6-PEX",
        "7-PE",
        "8-PEX_CTS_SDR",
    ]

    def __init__(self, _value=2):
        # type: (Union[str, int]) -> None
        super(PhHvacPipeMaterial, self).__init__(_value)

    def __eq__(self, other):
        # type: (PhHvacPipeMaterial) -> bool
        return self.value == other.value

    def __ne__(self, other):
        # type: (PhHvacPipeMaterial) -> bool
        return self.value != other.value

    def __hash__(self):
        # type: () -> int
        return hash(self.value)


# -- Piping -------------------------------------------------------------------


class PhHvacPipeSegment(_base._PhHVACBase):
    """A single pipe segment (linear) with geometry and a diameter

    Note: Following the LBT convention, while the geometry can be in a variety of
    units, thicknesses are all required to be in meters. This means that while the geometry
    will scale, the thickness and diameter will not.
    """

    # fmt: off
    def __init__(
        self,
        _geom,
        _diameter_mm=12.7,
        _insul_thickness_mm=12.7,
        _insul_conductivity=0.04,
        _insul_refl=True,
        _insul_quality=None,
        _daily_period=24,
        _water_temp_c=60.0,
        _material=2,
        *args,
        **kwargs
    ):  # fmt: on
        # type: (LineSegment3D, float, float, float, bool, None, float, float, int, *Any, **Any) -> None
        super(PhHvacPipeSegment, self).__init__()
        self.geometry = _geom
        self.diameter_mm = _diameter_mm
        self.insulation_thickness_mm = _insul_thickness_mm
        self.insulation_conductivity = _insul_conductivity
        self.insulation_reflective = _insul_refl
        self.insulation_quality = _insul_quality
        self.daily_period = _daily_period
        self.water_temp_c = _water_temp_c
        self.material = PhHvacPipeMaterial(_material)

    @property
    def length(self):
        # type: () -> float
        """Return the length of the pipe segment in model-units."""
        return self.geometry.length

    @property
    def diameter_m(self):
        # type: () -> float
        """Return the diameter of the pipe segment in meters."""
        return self.diameter_mm * 0.001

    @property
    def insulation_thickness_m(self):
        # type: () -> float
        """Return the insulation thickness of the pipe segment in meters."""
        return self.insulation_thickness_mm * 0.001

    def __copy__(self):
        # type: () -> PhHvacPipeSegment
        new_obj = PhHvacPipeSegment(self.geometry.duplicate())

        new_obj.diameter_mm = self.diameter_mm
        new_obj.material = PhHvacPipeMaterial(self.material.value)
        new_obj.insulation_thickness_mm = self.insulation_thickness_mm
        new_obj.insulation_conductivity = self.insulation_conductivity
        new_obj.insulation_reflective = self.insulation_reflective
        new_obj.insulation_quality = self.insulation_quality
        new_obj.daily_period = self.daily_period
        new_obj.water_temp_c = self.water_temp_c
        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = copy(self.user_data)

        return new_obj

    def duplicate(self):
        # type: () -> PhHvacPipeSegment
        return self.__copy__()

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Union[str, Dict]]
        d = super(PhHvacPipeSegment, self).to_dict()
        d["geometry"] = self.geometry.to_dict()
        d["diameter_mm"] = self.diameter_mm
        d["material_value"] = self.material.value
        d["insulation_thickness_mm"] = self.insulation_thickness_mm
        d["insulation_conductivity"] = self.insulation_conductivity
        d["insulation_reflective"] = self.insulation_reflective
        d["insulation_quality"] = self.insulation_quality
        d["daily_period"] = self.daily_period
        d["water_temp_c"] = self.water_temp_c

        if _include_properties:
            d["length"] = self.length

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> PhHvacPipeSegment
        new_obj = cls(_geom=LineSegment3D.from_dict(_input_dict["geometry"]))
        new_obj.diameter_mm = _input_dict["diameter_mm"]
        new_obj.material = PhHvacPipeMaterial(_input_dict["material_value"])
        new_obj.insulation_thickness_mm = _input_dict["insulation_thickness_mm"]
        new_obj.insulation_conductivity = _input_dict["insulation_conductivity"]
        new_obj.insulation_reflective = _input_dict["insulation_reflective"]
        new_obj.insulation_quality = _input_dict["insulation_quality"]
        new_obj.daily_period = _input_dict["daily_period"]
        new_obj.water_temp_c = _input_dict["water_temp_c"]
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict["user_data"]

        return new_obj

    def __str__(self):
        return "{}: diam={:.3f} (MM), length={:.3f} (Model-Unit)".format(
            self.__class__.__name__, self.diameter_mm, self.length
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return self.__repr__()

    def move(self, moving_vec3D):
        # type: (Vector3D) -> PhHvacPipeSegment
        """Move the pipe's geometry along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        Returns:
            A new PhHvacPipeSegment with the moved geometry.
        """
        dup = self.duplicate()
        dup.geometry = self.geometry.move(moving_vec3D)
        return dup

    def rotate(self, axis_3D, angle_degrees, origin_pt3D):
        # type: (Vector3D, float, Point3D) -> PhHvacPipeSegment
        """Rotate the pipe's geometry by a certain angle_degrees around an axis_3D and origin_pt3D.

        Right hand rule applies:
        If axis_3D has a positive orientation, rotation will be clockwise.
        If axis_3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_3D: A Vector3D axis_3D representing the axis_3D of rotation.
            angle_degrees: An angle_degrees for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeSegment with the rotated geometry.
        """
        dup = self.duplicate()
        dup.geometry = self.geometry.rotate(axis_3D, radians(angle_degrees), origin_pt3D)
        return dup

    def rotate_xy(self, angle_degrees, origin_pt3D):
        # type: (float, Point3D) -> PhHvacPipeSegment
        """Rotate the pipe's geometry counterclockwise in the XY plane by a certain angle_degrees.

        Args:
            angle_degrees: An angle_degrees in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeSegment with the rotated geometry.
        """
        dup = self.duplicate()
        dup.geometry = self.geometry.rotate_xy(radians(angle_degrees), origin_pt3D)
        return dup

    def reflect(self, normal_vec3D, origin_pt3D):
        # type: (Vector3D, Point3D) -> PhHvacPipeSegment
        """Reflected the pipe's geometry across a plane with the input normal vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        Returns:
            A new PhHvacPipeSegment with the reflected geometry.
        """
        dup = self.duplicate()
        dup.geometry = self.geometry.reflect(normal_vec3D, origin_pt3D)
        return dup

    def scale(self, scale_factor, origin_pt3D=None):
        # type: (float, Union[None, Point3D]) -> PhHvacPipeSegment
        """Scale the pipe's geometry by a factor from an origin_pt3D point.

        Note that following the LBT convention, while the geometry can be in a variety of
        units, thicknesses are all required to be in meters. This means that while the geometry
        will scale, the thickness and diameter will not.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        """
        new_pipe_segment = self.duplicate()
        new_pipe_segment.geometry = self.geometry.scale(scale_factor, origin_pt3D)
        return new_pipe_segment


class PhHvacPipeElement(_base._PhHVACBase):
    """A Pipe Element (Fixture) made up of one or more individual Pipe Segments."""

    def __init__(self):
        super(PhHvacPipeElement, self).__init__()
        self._segments = {}  # type: Dict[str, PhHvacPipeSegment]

    @property
    def segments(self):
        # type: () -> List[PhHvacPipeSegment]
        """Return a list of a;; the Pipe-Segments in the Pipe-Element."""
        return list(self._segments.values())

    @property
    def length(self):
        # type: () -> float
        """Return the total length of the pipe element in model-units."""
        return sum(s.length for s in self.segments)

    @property
    def diameter_mm(self):
        # type: () -> float
        """Return the length-weighted average diameter of all the pipe segments"""
        try:
            return sum(s.length * s.diameter_mm for s in self.segments) / self.length
        except ZeroDivisionError:
            return 0

    @property
    def water_temp_c(self):
        # type: () -> float
        """Return the length-weighted average water temperature of all the pipe segments"""
        try:
            return sum(s.length * s.water_temp_c for s in self.segments) / self.length
        except ZeroDivisionError:
            return 60.0

    @property
    def daily_period(self):
        # type: () -> float
        """Return the length-weighted average daily period of all the pipe segments"""
        try:
            return sum(s.length * s.daily_period for s in self.segments) / self.length
        except ZeroDivisionError:
            return 24.0

    @property
    def segment_names(self):
        # type: () -> List[str]
        """Return a list of the names of all the PipeSegments in the PipeElement."""
        return [s.display_name for s in self.segments]

    @property
    def material_name(self):
        # type: () -> str
        """Return the material name of the pipe element."""
        materials = {s.material for s in self.segments}
        if len(materials) == 0:
            return PhHvacPipeMaterial.allowed[0]
        elif len(materials) == 1:
            mat = materials.pop()
            return mat.value
        else:
            raise ValueError("Pipe segments: {} have different materials.".format(self.segment_names))

    def add_segment(self, _segment):
        # type: (PhHvacPipeSegment) -> None
        """Add a new Pipe Segment to the Pipe Element."""
        self._segments[_segment.identifier] = _segment

    def clear_segments(self):
        # type: () -> None
        """Clear all the segments from the pipe element."""
        self._segments = {}

    def __copy__(self):
        # type: () -> PhHvacPipeElement
        """Duplicate the pipe element."""
        new_obj = PhHvacPipeElement()

        for segment in self.segments:
            new_obj.add_segment(segment.duplicate())

        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data

        return new_obj

    def duplicate(self):
        # type: () -> PhHvacPipeElement
        return self.__copy__()

    def to_dict(self, _include_properties=False):
        # type: (bool) -> dict[str, Union[str, dict]]
        d = super(PhHvacPipeElement, self).to_dict()
        d["segments"] = {}
        for segment in self.segments:
            d["segments"][segment.identifier] = segment.to_dict(_include_properties)

        if _include_properties:
            d["length"] = self.length
            d["water_temp"] = self.water_temp_c
            d["daily_period"] = self.daily_period
            d["material_name"] = self.material_name
            d["diameter"] = self.diameter_mm

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> PhHvacPipeElement
        new_obj = cls()

        for seg_dict in _input_dict["segments"].values():
            new_obj._segments[seg_dict["identifier"]] = PhHvacPipeSegment.from_dict(seg_dict)
        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict["user_data"]

        return new_obj

    def __str__(self):
        return "{}: (display_name={}, identifier={} ) [{} segments, len={:.3f}]".format(
            self.__class__.__name__,
            self.display_name,
            self.identifier_short,
            len(self.segments),
            self.length,
        )

    def __repr__(self):
        return str(self)

    def ToString(self):
        return self.__repr__()

    def move(self, moving_vec3D):
        # type: (Vector3D) -> PhHvacPipeElement
        """Move the pipe's segments along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        Returns:
            A new PhHvacPipeElement with the moved segments.
        """
        new_pipe_element = self.duplicate()
        new_pipe_element.clear_segments()
        for segment in self.segments:
            new_pipe_element.add_segment(segment.move(moving_vec3D))
        return new_pipe_element

    def rotate(self, axis_3D, angle_degrees, origin_pt3D):
        # type: (Vector3D, float, Point3D) -> PhHvacPipeElement
        """Rotate the pipe's segments by a certain angle_degrees around an axis_3D and origin_pt3D.

        Right hand rule applies:
        If axis_3D has a positive orientation, rotation will be clockwise.
        If axis_3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_3D: A Vector3D axis_3D representing the axis_3D of rotation.
            angle_degrees: An angle_degrees for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeElement with the rotated segments.
        """
        new_pipe_element = self.duplicate()
        new_pipe_element.clear_segments()
        for segment in self.segments:
            new_pipe_element.add_segment(segment.rotate(axis_3D, angle_degrees, origin_pt3D))
        return new_pipe_element

    def rotate_xy(self, angle_degrees, origin_pt3D):
        # type: (float, Point3D) -> PhHvacPipeElement
        """Rotate the pipe's segments counterclockwise in the XY plane by a certain angle_degrees.

        Args:
            angle_degrees: An angle_degrees in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeElement with the rotated segments.
        """
        new_pipe_element = self.duplicate()
        new_pipe_element.clear_segments()
        for segment in self.segments:
            new_pipe_element.add_segment(segment.rotate_xy(angle_degrees, origin_pt3D))
        return new_pipe_element

    def reflect(self, normal_vec3D, origin_pt3D):
        # type: (Vector3D, Point3D) -> PhHvacPipeElement
        """Reflected the pipe's segments across a plane with the input normal_vec3D vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal_vec3D vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        Returns:
            A new PhHvacPipeElement with the reflected segments.
        """
        new_pipe_element = self.duplicate()
        new_pipe_element.clear_segments()
        for segment in self.segments:
            new_pipe_element.add_segment(segment.reflect(normal_vec3D, origin_pt3D))
        return new_pipe_element

    def scale(self, factor, origin_pt3D=None):
        # type: (float, Optional[Point3D]) -> PhHvacPipeElement
        """Scale the pipe's segments by a factor from an origin_pt3D point.

        Args:
            factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        Returns:
            A new PhHvacPipeElement with the scaled segments.
        """
        new_pipe_element = self.duplicate()
        new_pipe_element.clear_segments()
        for segment in self.segments:
            new_pipe_element.add_segment(segment.scale(factor, origin_pt3D))
        return new_pipe_element


class PhHvacPipeBranch(_base._PhHVACBase):
    """A 'Branch' Pipe which has geometry, and serves one or more 'Fixture' (Twig) pipe elements."""

    def __init__(self):
        # type: () -> None
        super(PhHvacPipeBranch, self).__init__()
        self.pipe_element = PhHvacPipeElement()
        self.fixtures = []  # type: (List[PhHvacPipeElement])

    @property
    def material_name(self):
        # type: () -> str
        """Return the material name of the pipe element."""
        return self.pipe_element.material_name

    @property
    def diameter_mm(self):
        # type: () -> float
        """Return the length-weighted diameter (MM) of the pipe element."""
        return self.pipe_element.diameter_mm

    @property
    def twigs(self):
        # type () -> List[PhPipeElement]
        """Alias for the 'fixtures' to better match Phius terminology."""
        return self.fixtures

    @property
    def segments(self):
        # type () -> List[PhPipeSegment]
        """Return a list of all the Pipe-Segments in the Branch."""
        return self.pipe_element.segments

    @property
    def length(self):
        # type: () -> float
        """Return the total length of the branch itself in model-units.
        For the total length of the Branch PLUS all fixtures, use 'total_length'.
        """
        return float(self.pipe_element.length)

    @property
    def water_temp_c(self):
        # type: () -> float
        """Return the length-weighted average water temperature of all the pipe segments."""
        return self.pipe_element.water_temp_c

    @property
    def daily_period(self):
        # type: () -> float
        """Return the length-weighted average daily period of all the pipe segments."""
        return self.pipe_element.daily_period

    @property
    def num_fixtures(self):
        # type: () -> int
        """Return the number of fixtures connected to the branch."""
        return len(self.fixtures)

    @property
    def total_length(self):
        # type: () -> float
        """Return the total length of the branch PLUS all fixture pipes in model-units."""
        return self.length + sum(f.length for f in self.fixtures)

    @property
    def total_home_run_fixture_length(self):
        # type: () -> float
        """Return the total length (in model-units) of all fixture pipes as measured from end to end.

        NOTE: This method will include the branch's length for EACH of
        the fixture pipes. The result will be a total hot-water transport length
        as if all the pipes were 'home-run' style. This value is used for the
        PHPP calculations and is not a true representation of the piping in the
        model.
        """
        return sum(fixture.length + self.length for fixture in self.fixtures)

    def add_fixture(self, _fixture):
        # type: (PhHvacPipeElement) -> None
        """Add a new HBPH Fixture (twig) PhPipeBranch to the Trunk."""
        self.fixtures.append(_fixture)

    def __copy__(self):
        # type: () -> PhHvacPipeBranch
        new_obj = PhHvacPipeBranch()

        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.pipe_element = self.pipe_element.duplicate()
        for fixture in self.fixtures:
            new_obj.add_fixture(fixture.duplicate())

        return new_obj

    def duplicate(self):
        # type: () -> PhHvacPipeBranch
        return self.__copy__()

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Union[str, Dict]]
        d = super(PhHvacPipeBranch, self).to_dict()
        d["pipe_element"] = self.pipe_element.to_dict(_include_properties)
        d["fixtures"] = {}
        for branch in self.fixtures:
            d["fixtures"][branch.identifier] = branch.to_dict(_include_properties)

        if _include_properties:
            d["length"] = self.length
            d["water_temp_c"] = self.water_temp_c
            d["daily_period"] = self.daily_period
            d["num_fixtures"] = self.num_fixtures
            d["total_length"] = self.total_length
            d["total_home_run_fixture_length"] = self.total_home_run_fixture_length
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> PhHvacPipeBranch
        new_obj = cls()

        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict["user_data"]
        new_obj.pipe_element = PhHvacPipeElement.from_dict(_input_dict["pipe_element"])
        for fixture_dict in _input_dict["fixtures"].values():
            new_obj.add_fixture(PhHvacPipeElement.from_dict(fixture_dict))

        return new_obj

    def __str__(self):
        # type: () -> str
        return "{}: (display_name={}, identifier={} ) [{} segments, len={:.1f}, {} fixtures connected]".format(
            self.__class__.__name__,
            self.display_name,
            self.identifier_short,
            len(self.segments),
            float(self.length),
            len(self.fixtures),
        )

    def __repr__(self):
        # type: () -> str
        return str(self)

    def ToString(self):
        # type: () -> str
        return self.__repr__()

    def move(self, moving_vec3D):
        # type: (Vector3D) -> PhHvacPipeBranch
        """Move the pipe's elements along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        Returns:
            A new PhHvacPipeBranch with the moved elements.
        """
        new_branch = self.duplicate()
        new_branch.fixtures = [fixture.move(moving_vec3D) for fixture in self.fixtures]
        new_branch.pipe_element = self.pipe_element.move(moving_vec3D)
        return new_branch

    def rotate(self, axis_3D, angle_degrees, origin_pt3D):
        # type: (Vector3D, float, Point3D) -> PhHvacPipeBranch
        """Rotate the pipe's elements by a certain angle_degrees around an axis_3D and origin_pt3D.

        Right hand rule applies:
        If axis_3D has a positive orientation, rotation will be clockwise.
        If axis_3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_3D: A Vector3D axis_3D representing the axis_3D of rotation.
            angle_degrees: An angle_degrees for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeBranch with the rotated elements.
        """
        new_branch = self.duplicate()
        new_branch.fixtures = [fixture.rotate(axis_3D, angle_degrees, origin_pt3D) for fixture in self.fixtures]
        new_branch.pipe_element = self.pipe_element.rotate(axis_3D, angle_degrees, origin_pt3D)
        return new_branch

    def rotate_xy(self, angle_degrees, origin_pt3D):
        # type: (float, Point3D) -> PhHvacPipeBranch
        """Rotate the pipe's elements counterclockwise in the XY plane by a certain angle_degrees.

        Args:
            angle_degrees: An angle_degrees in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeBranch with the rotated elements.
        """
        new_branch = self.duplicate()
        new_branch.fixtures = [fixture.rotate_xy(angle_degrees, origin_pt3D) for fixture in self.fixtures]
        new_branch.pipe_element = self.pipe_element.rotate_xy(angle_degrees, origin_pt3D)
        return new_branch

    def reflect(self, normal_vec3D, origin_pt3D):
        # type: (Vector3D, Point3D) -> PhHvacPipeBranch
        """Reflected the pipe's elements across a plane with the input normal_vec3D vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal_vec3D vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        Returns:
            A new PhHvacPipeBranch with the reflected elements.
        """
        new_branch = self.duplicate()
        new_branch.fixtures = [fixture.reflect(normal_vec3D, origin_pt3D) for fixture in self.fixtures]
        new_branch.pipe_element = self.pipe_element.reflect(normal_vec3D, origin_pt3D)
        return new_branch

    def scale(self, factor, origin_pt3D=None):
        # type: (float, Optional[Point3D]) -> PhHvacPipeBranch
        """Scale the pipe's elements by a factor from an origin_pt3D point.

        Args:
            factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        Returns:
            A new PhHvacPipeBranch with the scaled elements.
        """
        new_branch = self.duplicate()
        new_branch.fixtures = [fixture.scale(factor, origin_pt3D) for fixture in self.fixtures]
        new_branch.pipe_element = self.pipe_element.scale(factor, origin_pt3D)
        return new_branch


class PhHvacPipeTrunk(_base._PhHVACBase):
    """A 'Trunk' Pipe which has geometry, and serves one or more 'Branches'."""

    def __init__(self):
        # type: () -> None
        super(PhHvacPipeTrunk, self).__init__()
        self.pipe_element = PhHvacPipeElement()
        self.multiplier = 1  # type: int
        self.branches = []  # type: (List[PhHvacPipeBranch])
        self.demand_recirculation = False  # type: bool

    @property
    def material_name(self):
        # type: () -> str
        """Return the material name of the pipe element."""
        return self.pipe_element.material_name

    @property
    def diameter_mm(self):
        # type: () -> float
        """Return the length-weighted diameter (MM) name of the pipe element."""
        return self.pipe_element.diameter_mm

    @property
    def segments(self):
        # type () -> List[PhPipeSegment]
        """Return a list of all the Pipe-Segments in the Trunk."""
        return self.pipe_element.segments

    @property
    def length(self):
        # type: () -> float
        """Return the total length of the trunk itself in model-units."""
        return self.pipe_element.length

    @property
    def water_temp_c(self):
        # type: () -> float
        """Return the length-weighted average water temperature (deg-C) of all the pipe segments."""
        return self.pipe_element.water_temp_c

    @property
    def daily_period(self):
        # type: () -> float
        """Return the length-weighted average daily period of all the pipe segments."""
        return self.pipe_element.daily_period

    @property
    def num_fixtures(self):
        # type: () -> int
        """Return the number of fixtures connected to the trunk."""
        return sum(branch.num_fixtures for branch in self.branches)

    @property
    def total_length(self):
        # type: () -> float
        """Return the total length (in model-units) of the trunk PLUS all branches and fixture pipes in model-units."""
        return self.length + sum(branch.total_length for branch in self.branches)

    @property
    def total_home_run_fixture_length(self):
        # type: () -> float
        """Return the total length (in model-units) of all fixture pipes as measured from end to end.

        NOTE: This method will include the trunk's and branch's length for EACH of
        the fixture pipes. The result will be a total hot-water transport length
        as if all the pipes were 'home-run' style. This value is used for the
        PHPP calculations and is not a true representation of the piping in the
        model.
        """
        return sum(self.length + branch.total_home_run_fixture_length for branch in self.branches)

    def add_branch(self, _branch):
        # type: (PhHvacPipeBranch) -> None
        """Add a new HBPH PhPipeBranch to the Trunk."""
        self.branches.append(_branch)

    def __copy__(self):
        # type: () -> PhHvacPipeTrunk
        new_obj = PhHvacPipeTrunk()

        new_obj.identifier = self.identifier
        new_obj.display_name = self.display_name
        new_obj.user_data = self.user_data
        new_obj.pipe_element = self.pipe_element.duplicate()
        new_obj.multiplier = self.multiplier
        new_obj.demand_recirculation = self.demand_recirculation

        for branch in self.branches:
            new_obj.add_branch(branch.duplicate())

        return new_obj

    def duplicate(self):
        # type: () -> PhHvacPipeTrunk
        return self.__copy__()

    def to_dict(self, _include_properties=False):
        # type: (bool) -> Dict[str, Union[str, Dict]]
        d = super(PhHvacPipeTrunk, self).to_dict()
        d["pipe_element"] = self.pipe_element.to_dict(_include_properties)
        d["multiplier"] = self.multiplier
        d["demand_recirculation"] = self.demand_recirculation
        d["branches"] = {}
        for branch in self.branches:
            d["branches"][branch.identifier] = branch.to_dict(_include_properties)

        if _include_properties:
            d["length"] = self.length
            d["water_temp_c"] = self.water_temp_c
            d["daily_period"] = self.daily_period
            d["num_fixtures"] = self.num_fixtures
            d["total_length"] = self.total_length
            d["total_home_run_fixture_length"] = self.total_home_run_fixture_length
        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> PhHvacPipeTrunk
        new_obj = cls()

        new_obj.identifier = _input_dict["identifier"]
        new_obj.display_name = _input_dict["display_name"]
        new_obj.user_data = _input_dict["user_data"]
        new_obj.pipe_element = PhHvacPipeElement.from_dict(_input_dict["pipe_element"])
        new_obj.multiplier = _input_dict["multiplier"]
        new_obj.demand_recirculation = _input_dict.get("demand_recirculation", False)
        for branch_dict in _input_dict["branches"].values():
            new_obj.add_branch(PhHvacPipeBranch.from_dict(branch_dict))

        return new_obj

    def __str__(self):
        # type: () -> str
        return "{}: (display_name={}, identifier={}, demand_recirculation={}, multiplier={}) [{} segments, len={:.1f}, {} branches connected]".format(
            self.__class__.__name__,
            self.display_name,
            self.identifier_short,
            self.demand_recirculation,
            self.multiplier,
            len(self.segments),
            float(self.length),
            len(self.branches),
        )

    def __repr__(self):
        # type: () -> str
        return str(self)

    def ToString(self):
        # type: () -> str
        return self.__repr__()

    def move(self, moving_vec3D):
        # type: (Vector3D) -> PhHvacPipeTrunk
        """Move the pipe's elements along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        Returns:
            A new PhHvacPipeTrunk with the moved elements.
        """
        new_trunk = self.duplicate()
        new_trunk.branches = [branch.move(moving_vec3D) for branch in self.branches]
        new_trunk.pipe_element = self.pipe_element.move(moving_vec3D)
        return new_trunk

    def rotate(self, axis_3D, angle_degrees, origin_pt3D):
        # type: (Vector3D, float, Point3D) -> PhHvacPipeTrunk
        """Rotate the pipe's elements by a certain angle_degrees around an axis_3D and origin_pt3D.

        Right hand rule applies:
        If axis_3D has a positive orientation, rotation will be clockwise.
        If axis_3D has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_3D: A Vector3D axis_3D representing the axis_3D of rotation.
            angle_degrees: An angle_degrees for rotation in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeTrunk with the rotated elements.
        """
        new_trunk = self.duplicate()
        new_trunk.branches = [branch.rotate(axis_3D, angle_degrees, origin_pt3D) for branch in self.branches]
        new_trunk.pipe_element = self.pipe_element.rotate(axis_3D, angle_degrees, origin_pt3D)
        return new_trunk

    def rotate_xy(self, angle_degrees, origin_pt3D):
        # type: (float, Point3D) -> PhHvacPipeTrunk
        """Rotate the pipe's elements counterclockwise in the XY plane by a certain angle_degrees.

        Args:
            angle_degrees: An angle_degrees in degrees.
            origin_pt3D: A Point3D for the origin_pt3D around which the object will be rotated.
        Returns:
            A new PhHvacPipeTrunk with the rotated elements.
        """
        new_trunk = self.duplicate()
        new_trunk.branches = [branch.rotate_xy(angle_degrees, origin_pt3D) for branch in self.branches]
        new_trunk.pipe_element = self.pipe_element.rotate_xy(angle_degrees, origin_pt3D)
        return new_trunk

    def reflect(self, normal_vec3D, origin_pt3D):
        # type (Vector3D, Point3D) -> PhHvacPipeTrunk
        """Reflected the pipe's elements across a plane with the input normal_vec3D vector and origin_pt3D.

        Args:
            normal_vec3D: A Vector3D representing the normal_vec3D vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin_pt3D from which to reflect.
        Returns:
            A new PhHvacPipeTrunk with the reflected elements.
        """
        new_trunk = self.duplicate()
        new_trunk.branches = [branch.reflect(normal_vec3D, origin_pt3D) for branch in self.branches]
        new_trunk.pipe_element = self.pipe_element.reflect(normal_vec3D, origin_pt3D)
        return new_trunk

    def scale(self, factor, origin_pt3D=None):
        # type: (float, Optional[Point3D]) -> PhHvacPipeTrunk
        """Scale the pipe's elements by a factor from an origin_pt3D point.

        Args:
            factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin_pt3D from which to scale.
                If None, it will be scaled from the World origin_pt3D (0, 0, 0).
        Returns:
            A new PhHvacPipeTrunk with the scaled elements.
        """
        new_trunk = self.duplicate()
        new_trunk.branches = [branch.scale(factor, origin_pt3D) for branch in self.branches]
        new_trunk.pipe_element = self.pipe_element.scale(factor, origin_pt3D)
        return new_trunk



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Honeybee-PH-HVAC: Hot Water System."""

try:
    pass
except:
    pass  # IronPython

from uuid import uuid4

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry", e)

try:
    from honeybee_phhvac import hot_water_devices as hwd
    from honeybee_phhvac import hot_water_piping as hwp
except ImportError as e:
    raise ImportError("Failed to import honeybee_phhvac", e)


class PhHotWaterSystem_FromDictError(Exception):
    def __init__(self, _expected_types, _input_type):
        self.msg = 'Error: Expected type of "{}". Got: {}'.format(_expected_types, _input_type)
        super(PhHotWaterSystem_FromDictError, self).__init__(self.msg)


class PhHotWaterSystem(object):
    """PH-HVAC: Hot Water System."""

    def __init__(self):
        self.identifier = str(uuid4())
        self.id_num = 0
        self.display_name = "_unnamed_hot_water_system_"

        self.tank_1 = None  # type: Optional[hwd.PhHvacHotWaterTank]
        self.tank_2 = None  # type: Optional[hwd.PhHvacHotWaterTank]
        self.tank_buffer = None  # type: Optional[hwd.PhHvacHotWaterTank]
        self.tank_solar = None  # type: Optional[hwd.PhHvacHotWaterTank]

        self._heaters = {}  # type: Dict[str, hwd.PhHvacHotWaterHeater]
        self._distribution_piping = {}  # type: Dict[str, hwp.PhHvacPipeTrunk]
        self._recirc_piping = {}  # type: Dict[str, hwp.PhHvacPipeElement]

        self._number_tap_points = None  # type: Optional[int]

    @property
    def total_distribution_pipe_length(self):
        # type: () -> float
        """Returns the total length of all trunk, branch, and fixture piping."""
        return sum(_.total_length for _ in self._distribution_piping.values())

    @property
    def total_home_run_fixture_pipe_length(self):
        # type: () -> float
        """Returns the total length of all fixture piping from end to end.

        NOTE: This method will count the branch and trunk lengths for EACH of
        the fixture pipes. The result will be a total hot-water transport length
        as if all the pipes were 'home-run' style. This value is used for the
        PHPP calculations and is not a true representation of the piping in the
        model.
        """
        return sum(_.total_home_run_fixture_length for _ in self._distribution_piping.values())

    @property
    def total_recirc_pipe_length(self):
        # type: () -> float
        """Returns the total length of all recirculation piping."""
        return sum(_.length for _ in self._recirc_piping.values())

    @property
    def recirc_temp(self):
        # type: () -> float
        """Return the length weighted average of recirculation piping temperatures"""
        if not self._recirc_piping or self.total_recirc_pipe_length == 0:
            return 60.0
        return sum([v.water_temp_c * v.length for v in self._recirc_piping.values()]) / self.total_recirc_pipe_length

    @property
    def recirc_hours(self):
        # type: () -> int
        """Return the length-weighted average of recirculation piping hours."""
        if not self._recirc_piping or self.total_recirc_pipe_length == 0:
            return 24
        return int(
            sum([v.daily_period * v.length for v in self._recirc_piping.values()]) / self.total_recirc_pipe_length
        )

    @property
    def number_tap_points(self):
        # type: () -> int
        """Unless set explicitly by the user, will return the number of Branch Pipe Elements."""
        if self._number_tap_points:
            return self._number_tap_points
        else:
            return sum(br.num_fixtures for br in self._distribution_piping.values())

    @number_tap_points.setter
    def number_tap_points(self, _input):
        # type: (Optional[int]) -> None
        if _input:
            self._number_tap_points = int(_input)

    @property
    def heaters(self):
        # type: () -> ValuesView[hwd.PhHvacHotWaterHeater]
        """Returns a list of all the heaters on the system."""
        return self._heaters.values()

    def clear_heaters(self):
        self._heaters = {}

    def add_heater(self, _heater):
        # type: (Optional[hwd.PhHvacHotWaterHeater]) -> None
        """Adds a new hot-water heater to the system."""
        if not _heater:
            return

        assert hasattr(_heater, "to_dict"), 'Error: HW-Heater "{}" is not serializable?'.format(_heater)
        self._heaters[_heater.identifier] = _heater

    def add_distribution_piping(self, _distribution_piping, _key=None):
        # type: (Union[hwp.PhHvacPipeTrunk, hwp.PhHvacPipeBranch, hwp.PhHvacPipeElement], Optional[str]) -> None
        """Add a new distribution (branch, trunk, fixture) to the system.

        If a branch or fixture pipe is passed, a 0-length trunk will be created and the
        branch or fixture will be added to it before adding to the system.
        """

        if isinstance(_distribution_piping, hwp.PhHvacPipeTrunk):
            # -- Add the trunk to the collection
            new_trunk = _distribution_piping
        elif isinstance(_distribution_piping, hwp.PhHvacPipeBranch):
            # -- Build a new Trunk and add the branch to it
            new_trunk = hwp.PhHvacPipeTrunk()
            new_trunk.add_branch(_distribution_piping)
        elif isinstance(_distribution_piping, hwp.PhHvacPipeElement):
            # -- Build a new Trunk and Branch, add the fixture to it
            new_branch = hwp.PhHvacPipeBranch()
            new_branch.add_fixture(_distribution_piping)
            new_trunk = hwp.PhHvacPipeTrunk()
            new_trunk.add_branch(new_branch)
        else:
            raise ValueError(
                'Error: Expected type of "PhPipeTrunk", "PhPipeBranch", or "PhPipeElement". Got: {}'.format(
                    type(_distribution_piping)
                )
            )

        self._distribution_piping[_key or new_trunk.identifier] = new_trunk

    def clear_distribution_piping(self):
        """Clear all distribution piping (Trunks) from the system."""
        self._distribution_piping = {}

    @property
    def distribution_piping(self):
        # type: () -> ValuesView[hwp.PhHvacPipeTrunk]
        """Returns a list of all the distribution-piping (Trunks) in the system."""
        return self._distribution_piping.values()

    def add_recirc_piping(self, _recirc_piping, _key=None):
        # type: (hwp.PhHvacPipeElement, Optional[str]) -> None
        self._recirc_piping[_key or _recirc_piping.identifier] = _recirc_piping

    def clear_recirc_piping(self):
        self._recirc_piping = {}

    @property
    def recirc_piping(self):
        # type: () -> ValuesView[hwp.PhHvacPipeElement]
        """Returns a list of all the recirculation-piping objects in the system."""
        return self._recirc_piping.values()

    @property
    def tanks(self):
        # type: () -> list[hwd.PhHvacHotWaterTank | None]
        """Return a list of the system tanks in order (1, 2, buffer, solar)."""
        return [self.tank_1, self.tank_2, self.tank_buffer, self.tank_solar]

    def to_dict(self, abridged=False, _include_properties=False):
        # type: (bool, bool) -> dict[str, dict]
        d = {}
        if abridged:
            d["type"] = "PhHvacHotWaterSystemAbridged"
        else:
            d["type"] = "PhHvacHotWaterSystemPh"

        d["id_num"] = self.id_num
        d["identifier"] = self.identifier
        d["display_name"] = self.display_name
        if self.tank_1:
            d["tank_1"] = self.tank_1.to_dict(_include_properties)
        if self.tank_2:
            d["tank_2"] = self.tank_2.to_dict(_include_properties)
        if self.tank_buffer:
            d["tank_buffer"] = self.tank_buffer.to_dict(_include_properties)
        if self.tank_solar:
            d["tank_solar"] = self.tank_solar.to_dict(_include_properties)

        d["heaters"] = {}
        for heater in self.heaters:
            d["heaters"][heater.identifier] = heater.to_dict(_include_properties)

        d["distribution_piping"] = {}
        for distribution_piping in self.distribution_piping:
            d["distribution_piping"][distribution_piping.identifier] = distribution_piping.to_dict(_include_properties)

        d["recirc_piping"] = {}
        for recirc_piping in self.recirc_piping:
            d["recirc_piping"][recirc_piping.identifier] = recirc_piping.to_dict(_include_properties)

        d["number_tap_points"] = self._number_tap_points
        d["recirc_temp"] = self.recirc_temp
        d["recirc_hours"] = self.recirc_hours

        if _include_properties:
            d["total_distribution_pipe_length"] = self.total_distribution_pipe_length
            d["total_home_run_fixture_pipe_length"] = self.total_home_run_fixture_pipe_length
            d["total_recirc_pipe_length"] = self.total_recirc_pipe_length

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (dict) -> PhHotWaterSystem
        valid_types = ("PhHvacHotWaterSystemPh", "PhHvacHotWaterSystemAbridged")
        if _input_dict["type"] not in valid_types:
            raise PhHotWaterSystem_FromDictError(valid_types, _input_dict["type"])

        new_system = cls()
        new_system.identifier = _input_dict.get("identifier", str(uuid4()))
        new_system.id_num = _input_dict["id_num"]
        new_system.display_name = _input_dict["display_name"]

        if _input_dict.get("tank_1", None):
            new_system.tank_1 = hwd.PhHvacHotWaterTank.from_dict(_input_dict["tank_1"])
        if _input_dict.get("tank_2", None):
            new_system.tank_2 = hwd.PhHvacHotWaterTank.from_dict(_input_dict["tank_2"])
        if _input_dict.get("tank_buffer", None):
            new_system.tank_buffer = hwd.PhHvacHotWaterTank.from_dict(_input_dict["tank_buffer"])
        if _input_dict.get("tank_solar", None):
            new_system.tank_solar = hwd.PhHvacHotWaterTank.from_dict(_input_dict["tank_solar"])

        for heater_dict in _input_dict["heaters"].values():
            new_system.add_heater(hwd.PhHvacHotWaterHeaterBuilder.from_dict(heater_dict))

        for distribution_piping_dict in _input_dict["distribution_piping"].values():
            new_system.add_distribution_piping(hwp.PhHvacPipeTrunk.from_dict(distribution_piping_dict))

        for recirc_piping_dict in _input_dict["recirc_piping"].values():
            new_system.add_recirc_piping(hwp.PhHvacPipeElement.from_dict(recirc_piping_dict))

        new_system._number_tap_points = _input_dict["number_tap_points"]

        return new_system

    def apply_properties_from_dict(self, abridged_data):
        return

    def __copy__(self):
        # type: () -> PhHotWaterSystem
        new_obj = PhHotWaterSystem()
        new_obj.id_num = self.id_num
        new_obj.display_name = self.display_name

        if self.tank_1:
            new_obj.tank_1 = self.tank_1.duplicate()
        if self.tank_2:
            new_obj.tank_2 = self.tank_2.duplicate()
        if self.tank_buffer:
            new_obj.tank_buffer = self.tank_buffer.duplicate()
        if self.tank_solar:
            new_obj.tank_solar = self.tank_solar.duplicate()

        for k, v in self._heaters.items():
            new_obj.add_heater(v.duplicate())

        for k, v in self._distribution_piping.items():
            new_obj.add_distribution_piping(v.duplicate(), _key=k)

        for k, v in self._recirc_piping.items():
            new_obj.add_recirc_piping(v.duplicate(), _key=k)

        new_obj._number_tap_points = self._number_tap_points

        return new_obj

    def duplicate(self):
        # type: (Any) -> PhHotWaterSystem
        return self.__copy__()

    def __str__(self):
        return "{}: id={}".format(self.__class__.__name__, self.id_num)

    def __repr__(self):
        """Properties representation."""
        return "{!r}(id_num={!r}, tank_1={!r}, tank_2={!r}, tank_buffer={!r}, tank_solar={!r})".format(
            self.__class__.__name__,
            self.id_num,
            self.tank_1,
            self.tank_2,
            self.tank_buffer,
            self.tank_solar,
        )

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __add__(self, other):
        # type: (PhHotWaterSystem) -> PhHotWaterSystem
        new_obj = self.duplicate()

        for heater in self.heaters:
            new_obj.add_heater(heater)
        for heater in other.heaters:
            new_obj.add_heater(heater)

        for distribution_pipe in self.distribution_piping:
            new_obj.add_distribution_piping(distribution_pipe)
        for distribution_pipe in other.distribution_piping:
            new_obj.add_distribution_piping(distribution_pipe)

        for recirc_pipe in self.recirc_piping:
            new_obj.add_recirc_piping(recirc_pipe)
        for recirc_pipe in other.recirc_piping:
            new_obj.add_recirc_piping(recirc_pipe)

        if self._number_tap_points or other._number_tap_points:
            # -- If either have their tap-points number set explicitly by the use
            # -- set the new object # as well
            new_obj.number_tap_points = self.number_tap_points + other.number_tap_points
        else:
            # -- If neither obj's have their tap-point number set explicitly by the user
            # -- Set as None, which will return the length of the Branch Piping dict by default.
            self._number_tap_points = None

        return new_obj

    def __radd__(self, other):
        # type: (PhHotWaterSystem) -> PhHotWaterSystem
        if isinstance(other, int):
            return self
        else:
            return self + other

    def __eq__(self, other):
        # type: (PhHotWaterSystem) -> bool
        if not isinstance(other, PhHotWaterSystem):
            return False

        if self.id_num != other.id_num:
            return False

        if self.tank_1 != other.tank_1:
            return False
        if self.tank_2 != other.tank_2:
            return False
        if self.tank_buffer != other.tank_buffer:
            return False
        if self.tank_solar != other.tank_solar:
            return False

        if len(self.heaters) != len(other.heaters):
            return False
        for heater in self.heaters:
            if heater not in other.heaters:
                return False

        if len(self.distribution_piping) != len(other.distribution_piping):
            return False
        for distribution_pipe in self.distribution_piping:
            if distribution_pipe not in other.distribution_piping:
                return False

        if len(self.recirc_piping) != len(other.recirc_piping):
            return False
        for recirc_pipe in self.recirc_piping:
            if recirc_pipe not in other.recirc_piping:
                return False

        if self.number_tap_points != other.number_tap_points:
            return False

        return True

    def move(self, moving_vec3D):
        # type: (Point3D) -> PhHotWaterSystem
        """Move the System's piping along a vector.

        Args:
            moving_vec3D: A Vector3D with the direction and distance to move the ray.
        Reflect:
            A new PhHotWaterSystem object with the moved piping.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.clear_distribution_piping()
        new_system.clear_recirc_piping()

        for k, pipe in self._distribution_piping.items():
            new_system.add_distribution_piping(pipe.move(moving_vec3D), _key=k)

        for k, pipe in self._recirc_piping.items():
            new_system.add_recirc_piping(pipe.move(moving_vec3D), _key=k)

        return new_system

    def rotate(self, axis_vec3D, angle_degrees, origin_pt3D):
        # type: (Point3D, float, Point3D) -> PhHotWaterSystem
        """Rotate the System's piping by a certain angle around an axis and origin.

        Right hand rule applies:
        If axis has a positive orientation, rotation will be clockwise.
        If axis has a negative orientation, rotation will be counterclockwise.

        Args:
            axis_vec3D: A Vector3D axis representing the axis of rotation.
            angle_degrees: An angle for rotation in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        Returns:
            A new PhHotWaterSystem object with the rotated piping.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.clear_distribution_piping()
        new_system.clear_recirc_piping()

        for k, pipe in self._distribution_piping.items():
            new_system.add_distribution_piping(pipe.rotate(axis_vec3D, angle_degrees, origin_pt3D), _key=k)

        for k, pipe in self._recirc_piping.items():
            new_system.add_recirc_piping(pipe.rotate(axis_vec3D, angle_degrees, origin_pt3D), _key=k)

        return new_system

    def rotate_xy(self, angle_degrees, origin_pt3D):
        # type: (float, Point3D) -> PhHotWaterSystem
        """Rotate the System's piping counterclockwise in the XY plane by a certain angle.

        Args:
            angle_degrees: An angle in degrees.
            origin_pt3D: A Point3D for the origin around which the object will be rotated.
        Returns:
            A new PhHotWaterSystem object with the rotated piping.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.clear_distribution_piping()
        new_system.clear_recirc_piping()

        for k, pipe in self._distribution_piping.items():
            new_system.add_distribution_piping(pipe.rotate_xy(angle_degrees, origin_pt3D), _key=k)

        for k, pipe in self._recirc_piping.items():
            new_system.add_recirc_piping(pipe.rotate_xy(angle_degrees, origin_pt3D), _key=k)

        return new_system

    def reflect(self, normal_vec3D, origin_pt3D):
        # type: (Vector3D, Point3D) -> PhHotWaterSystem
        """Reflected the System's piping across a plane with the input normal vector and origin.

        Args:
            normal_vec3D: A Vector3D representing the normal vector for the plane across
                which the line segment will be reflected. THIS VECTOR MUST BE NORMALIZED.
            origin_pt3D: A Point3D representing the origin from which to reflect.
        Returns:
            A new PhHotWaterSystem object with the reflected piping.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.clear_distribution_piping()
        new_system.clear_recirc_piping()

        for k, pipe in self._distribution_piping.items():
            new_system.add_distribution_piping(pipe.reflect(normal_vec3D, origin_pt3D), _key=k)

        for k, pipe in self._recirc_piping.items():
            new_system.add_recirc_piping(pipe.reflect(normal_vec3D, origin_pt3D), _key=k)

        return new_system

    def scale(self, scale_factor, origin_pt3D=None):
        # type: (float, Optional[Point3D]) -> PhHotWaterSystem
        """Scale the System's piping by a factor from an origin point.

        Args:
            scale_factor: A number representing how much the line segment should be scaled.
            origin_pt3D: A Point3D representing the origin from which to scale.
                If None, it will be scaled from the World origin (0, 0, 0).
        Returns:
            A new PhHotWaterSystem object with the scaled piping.
        """
        new_system = self.duplicate()
        new_system.identifier = self.identifier
        new_system.clear_distribution_piping()
        new_system.clear_recirc_piping()

        for k, pipe in self._distribution_piping.items():
            new_system.add_distribution_piping(pipe.scale(scale_factor, origin_pt3D), _key=k)

        for k, pipe in self._recirc_piping.items():
            new_system.add_recirc_piping(pipe.scale(scale_factor, origin_pt3D), _key=k)

        return new_system



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""HB-Aperture Passive House (PH) Properties."""


try:
    from typing import Any, Dict, Optional
except ImportError:
    pass  # IronPython 2.7


class ShadingDimensions(object):
    """PHPP Style shading dimension info"""

    def __init__(self):
        # Horizon Shading
        self.h_hori = None  # type: Optional[float]
        self.d_hori = None  # type: Optional[float]

        # Side Reveal
        self.o_reveal = None  # type: Optional[float]
        self.d_reveal = None  # type: Optional[float]

        # Overhangs
        self.o_over = None  # type: Optional[float]
        self.d_over = None  # type: Optional[float]

    def __copy__(self, new_host=None):
        # type: (Any) -> ShadingDimensions

        new_obj = ShadingDimensions()

        new_obj.d_hori = self.d_hori
        new_obj.h_hori = self.h_hori
        new_obj.d_reveal = self.d_reveal
        new_obj.o_reveal = self.o_reveal
        new_obj.d_over = self.d_over
        new_obj.o_over = self.o_over

        return new_obj

    def duplicate(self, new_host=None):
        # type: (Any) -> ShadingDimensions
        return self.__copy__(new_host=new_host)

    def to_dict(self):
        # type: () -> Dict[str, Any]
        d = {}

        d["d_hori"] = self.d_hori
        d["h_hori"] = self.h_hori
        d["d_reveal"] = self.d_reveal
        d["o_reveal"] = self.o_reveal
        d["d_over"] = self.d_over
        d["o_over"] = self.o_over

        return d

    @classmethod
    def from_dict(cls, _input_dict):
        # type: (Dict) -> ShadingDimensions
        new_obj = cls()

        new_obj.d_hori = _input_dict["d_hori"]
        new_obj.h_hori = _input_dict["h_hori"]
        new_obj.d_reveal = _input_dict["d_reveal"]
        new_obj.o_reveal = _input_dict["o_reveal"]
        new_obj.d_over = _input_dict["d_over"]
        new_obj.o_over = _input_dict["o_over"]

        return new_obj


class AperturePhProperties(object):
    def __init__(self, _host):
        self._host = _host
        self.id_num = 0
        self.winter_shading_factor = 0.75
        self.summer_shading_factor = 0.75
        self.shading_dimensions = None  # type: Optional[ShadingDimensions]
        self.variant_type = "_unnamed_type_"
        self.install_depth = 0.1016  # m
        self.default_monthly_shading_correction_factor = 1.0

    @property
    def host(self):
        return self._host

    def duplicate(self, new_host=None):
        # type: (Any) -> AperturePhProperties
        _host = new_host or self._host
        new_properties_obj = AperturePhProperties(_host)
        new_properties_obj.id_num = self.id_num

        new_properties_obj.winter_shading_factor = self.winter_shading_factor
        new_properties_obj.summer_shading_factor = self.summer_shading_factor
        if self.shading_dimensions:
            new_properties_obj.shading_dimensions = self.shading_dimensions.duplicate(self)
        new_properties_obj.variant_type = self.variant_type
        new_properties_obj.install_depth = self.install_depth
        new_properties_obj.default_monthly_shading_correction_factor = self.default_monthly_shading_correction_factor

        return new_properties_obj

    def ToString(self):
        return self.__repr__()

    def __repr__(self):
        return "HB-Aperture Passive House Properties: [host: {}]".format(self.host.display_name)

    def to_dict(self, abridged=False):
        # type: (bool) -> Dict[str, Dict[str, Any]]
        d = {}
        d["type"] = "AperturePhProperties" if not abridged else "AperturePhPropertiesAbridged"
        d["id_num"] = self.id_num
        d["winter_shading_factor"] = self.winter_shading_factor
        d["summer_shading_factor"] = self.summer_shading_factor
        if self.shading_dimensions:
            d["shading_dims"] = self.shading_dimensions.to_dict()
        d["variant_type"] = self.variant_type
        d["install_depth"] = self.install_depth
        d["default_monthly_shading_correction_factor"] = self.default_monthly_shading_correction_factor

        return {"ph": d}

    @classmethod
    def from_dict(cls, _input_dict, host):
        # type: (Dict[str, Any], Any) -> AperturePhProperties
        assert _input_dict["type"] == "AperturePhProperties", "Expected AperturePhProperties. Got {}.".format(
            _input_dict["type"]
        )

        new_prop = cls(host)
        new_prop.id_num = _input_dict["id_num"]
        new_prop.winter_shading_factor = _input_dict["winter_shading_factor"]
        new_prop.summer_shading_factor = _input_dict["summer_shading_factor"]
        new_prop.variant_type = _input_dict["variant_type"]

        # Use get to ensure backwards compatibility for now
        new_prop.install_depth = _input_dict.get("install_depth", 0.1)

        shading_dim_dict = _input_dict.get("shading_dims", None)
        if shading_dim_dict:
            new_prop.shading_dimensions = ShadingDimensions.from_dict(shading_dim_dict)

        new_prop.default_monthly_shading_correction_factor = _input_dict.get(
            "default_monthly_shading_correction_factor", 1.0
        )

        return new_prop

    def apply_properties_from_dict(self, _aperture_prop_dict):
        # type: (Dict[str, Any]) -> None
        """Apply properties from an AperturePhPropertiesAbridged dictionary.

        Arguments:
        ----------
            * _aperture_prop_dict (dict): An AperturePhPropertiesAbridged dictionary loaded from
                the Aperture object itself. Unabridged.
        """

        self.winter_shading_factor = _aperture_prop_dict["winter_shading_factor"]
        self.summer_shading_factor = _aperture_prop_dict["summer_shading_factor"]
        self.variant_type = _aperture_prop_dict["variant_type"]

        # Use get to ensure backwards compatibility for now
        self.install_depth = _aperture_prop_dict.get("install_depth", 0.1016)  # default = 4in.

        shading_dim_dict = _aperture_prop_dict.get("shading_dims", None)
        if shading_dim_dict:
            self.shading_dimensions = ShadingDimensions.from_dict(shading_dim_dict)

        self.default_monthly_shading_correction_factor = _aperture_prop_dict.get(
            "default_monthly_shading_correction_factor", 1.0
        )

        return None



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Utility function for creating a simple Histrogram from a data set."""

try:
    pass
except:
    pass  # IronPython

from collections import defaultdict


def generate_histogram(_data, _num_bins):
    # type: (Collection[float], int) -> dict[int, dict[str, float]]
    """Creates a Histogram of input data, in n-bins.

    Arguments:
    ----------
        * _data (Collection[float]): Collection of numeric values to use as the data source.
        * _num_bins (int): Number of bins to split the input data set into.

    Returns:
    --------
        * dict[int, dict[str, float]]: ie:
            {
                0: {'average_value'=12.0, 'frequency'=0.25},
                1: {'average_value'=6.0, 'frequency'=0.34},
                2: ...
            },
    """

    # -- Bin the data
    binned_data = defaultdict(list)
    maximum = max(_data)
    minimum = min(_data)
    val_range = maximum - minimum

    if val_range == 0:
        binned_data[0] = _data
    else:
        for d in _data:
            normalized_value = (maximum - d) / val_range
            bin = round(normalized_value * (_num_bins - 1))
            binned_data[bin].append(d)

    # -- Format the data for output

    output = {}
    for k, v in binned_data.items():
        output[k] = {
            "average_value": (sum(v) / len(v)) if len(v) > 0 else 0,
            "frequency": len(v) / len(_data),
        }

    return output



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-


"""DEPRECATED in favor of the new 'honeybee_phhvac' module."""


from honeybee_energy_ph.hvac import _base


class Ventilator(_base._PhHVACBase):
    message = "Ventilator is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhVentilationSystem(_base._PhHVACBase):
    message = "PhVentilationSystem is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class _ExhaustVentilatorBase(_base._PhHVACBase):
    message = "_ExhaustVentilatorBase is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class ExhaustVentDryer(_ExhaustVentilatorBase):
    message = "ExhaustVentDryer is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class ExhaustVentKitchenHood(_ExhaustVentilatorBase):
    message = "ExhaustVentKitchenHood is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class ExhaustVentUserDefined(_ExhaustVentilatorBase):
    message = "ExhaustVentUserDefined is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)


class PhExhaustDeviceBuilder(object):
    message = "PhExhaustDeviceBuilder is deprecated, use 'honeybee_phhvac' instead"

    def __init__(self, *args, **kwargs):
        raise Exception(self.message)



# -*- coding: utf-8 -*-
# -*- Python Version: 2.7 -*-

"""Package: Honeybee-PH-Standards

These classes are designed to be used within a Honeybee-editing environment, such as Rhino or
Grasshopper. These modules include various program standards used for Passive House buildings.
"""

# load all functions that extends honeybee core library
# Make sure the load `honeybee_energy` before anything else to avoid import errors.
import honeybee_energy._extend_honeybee
